import re
import urllib.request as req
from .__proc__ import _duration
from .__hyp__ import _HyperThread



class _VideoBulk:

    def __init__(self, iterable:list):
        self._ls = iterable

    @property
    def title(self):

        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"\"title\":\"(.*?)\""
            return re.findall(pattern, raw)[0]

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def views(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"\"videoViewCountRenderer\":{\"viewCount\":{\"simpleText\":\"(.*?)\""
            views = re.findall(pattern, raw)
            return views[0][:-6] if len(views) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def likes(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"toggledText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?) "
            likes = re.findall(pattern, raw)
            return likes[0] if len(likes) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def dislikes(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"DISLIKE\"},\"defaultText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?) dislikes\""
            dislikes = re.findall(pattern, raw)
            return dislikes[0] if len(dislikes) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def duration(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"approxDurationMs\":\"(.*?)\""
            duration = re.findall(pattern, raw)
            return _duration(int(int(duration[0]) / 1000))

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def upload_date(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"uploadDate\":\"(.*?)\""
            return re.findall(pattern, raw)[0]

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def parent(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"channelIds\":\[\"(.*?)\""
            return re.findall(pattern, raw)[0]

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def description(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"shortDescription\":\"(.*)\",\"isCrawlable"
            return re.findall(pattern, raw)[0]

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def thumbnail(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"playerMicroformatRenderer\":{\"thumbnail\":{\"thumbnails\":\[{\"url\":\"(.*?)\""
            return re.findall(pattern, raw)[0]

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def tags(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"<meta name=\"keywords\" content=\"(.*?)\">"
            tags = re.findall(pattern, raw)
            return tags[0].split(',') if len(tags) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/watch?v={item}' for item in self._ls])


    @property
    def url(self):
        return [f'https://www.youtube.com/watch?v={item}' for item in self._ls]


    @property
    def id(self):
        return self._ls


import re
import urllib.request
from .__vid__ import Video
from .__proc__ import _filter


class Playlist:
    def __init__(self, playlist_id: str):

        """

        :param str playlist_id: the _id of the playlist

        """

        if 'youtube.com' in playlist_id:
            self.id = re.findall(r'=(.*)', playlist_id)[0]
        else:
            self.id = playlist_id


    @property
    def info(self):

        """

        :return: a dict containing playlist info

        dict = {
                'name': -> str,
                'url': -> str,
                'video_count': -> int,
                'videos': -> bulk,
                'thumbnail': -> str,
            }

        """

        url = f'https://www.youtube.com/playlist?list={self.id}'

        raw = urllib.request.urlopen(url).read().decode()

        name_data = re.findall(r"{\"title\":\"(.*?)\"", raw)

        video_count_data = re.findall(r"stats\":\[{\"runs\":\[{\"text\":\"(.*?)\"", raw)

        thumbnails = re.findall(r"og:image\" content=\"(.*?)\?", raw)

        data_dict = {

            'name': name_data[0] if len(name_data) != 0 else None,
            'url': f'https://www.youtube.com/playlist?list={self.id}',
            'video_count': video_count_data[0] if len(video_count_data) != 0 else None,
            'videos': _filter(re.findall(r"videoId\":\"(.*?)\"", raw)),
            'thumbnail': thumbnails[0] if len(thumbnails) != 0 else None,
        }

        return data_dict

    @property
    def name(self):

        """

        :return: the name of the playlist

        """

        url = f'https://www.youtube.com/playlist?list={self.id}'
        raw = urllib.request.urlopen(url).read().decode()
        name_data = re.findall(r"{\"title\":\"(.*?)\"", raw)
        return name_data[0] if len(name_data) != 0 else None

    @property
    def url(self):

        """

        :return: url of the playlist

        """

        url = f'https://www.youtube.com/playlist?list={self.id}'
        return url

    @property
    def video_count(self):

        """

        :return: total number of videos in that playlist

        """

        url = f'https://www.youtube.com/playlist?list={self.id}'
        raw = urllib.request.urlopen(url).read().decode()
        video_count = re.findall(r"stats\":\[{\"runs\":\[{\"text\":\"(.*?)\"", raw)
        return video_count[0] if len(video_count) != 0 else None

    def videos(self, limit: int = None):

        """

        :param int limit: number of videos the user want from the playlist
        :return: list of < video objects > for each video in the playlist (consider limit)

        """

        url = f'https://www.youtube.com/playlist?list={self.id}'
        raw = urllib.request.urlopen(url).read().decode()
        videos = re.findall(r"videoId\":\"(.*?)\"", raw)
        pure = _filter(limit=limit, iterable=videos)
        return [Video(item) for item in pure]

    @property
    def thumbnail(self):

        """

        :return: url of the thumbnail of the playlist

        """

        url = f'https://www.youtube.com/playlist?list={self.id}'
        raw = urllib.request.urlopen(url).read().decode()
        thumbnails = re.findall(r"og:image\" content=\"(.*?)\?", raw)
        return thumbnails[0] if len(thumbnails) != 0 else None


import re
import urllib.request as req
from .__hyp__ import _HyperThread



class _ChannelBulk:

    def __init__(self, iterable:list):
        self._ls = iterable

    @property
    def id(self):
        return self._ls


    @property
    def url(self):
        return [f'https://www.youtube.com/channel/{item}' for item in self._ls]


    @property
    def name(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"channelMetadataRenderer\":{\"title\":\"(.*?)\""
            return re.findall(pattern, raw)[0]

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def subscribers(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"\"subscriberCountText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?)\""
            temp = re.findall(pattern, raw)
            return temp[0][:-12] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def total_views(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"\"viewCountText\":{\"simpleText\":\"(.*?)\"}"
            temp = re.findall(pattern, raw)
            return temp[0][:-6] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def joined(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"{\"text\":\"Joined \"},{\"text\":\"(.*?)\"}"
            temp = re.findall(pattern, raw)
            return temp[0] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def country(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"\"country\":{\"simpleText\":\"(.*?)\"}"
            temp = re.findall(pattern, raw)
            return temp[0] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def custom_url(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"\"canonicalChannelUrl\":\"(.*?)\""
            ins = re.findall(pattern, raw)
            temp = ins[0] if len(ins) > 0 else None
            return temp if '/channel' not in temp and temp is not None else None


        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])

    @property
    def description(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"{\"description\":{\"simpleText\":\"(.*?)\"}"
            temp = re.findall(pattern, raw)
            return temp[0] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def avatar_url(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = "height\":88},{\"url\":\"(.*?)\""
            temp = re.findall(pattern, raw)
            return temp[0] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def banner_url(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"width\":1280,\"height\":351},{\"url\":\"(.*?)\""
            temp = re.findall(pattern, raw)
            return temp[0] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


    @property
    def verified(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            isVerified = re.search(r'label":"Verified', raw)
            return True if isVerified else False

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])

    @property
    def live(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            isLive = re.search(r'{"text":" watching"}', raw)
            return True if isLive else False

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/channel/{item}/about' for item in self._ls])


import re
import urllib.request
from .__vid__ import Video
from .__proc__ import _filter
from .__vidbulk__ import _VideoBulk


class Extras:

    def __init__(self):
        pass



    @property
    def Trending(self):

        """

        :return: < video object > of #1 on trending video

        """

        url = f'https://www.youtube.com/feed/trending'
        raw = urllib.request.urlopen(url).read().decode()
        data = re.findall(r"videoId\":\"(.*?)\"", raw)
        return Video(data[0]) if len(data) != 0 else None


    @property
    def Music(self):

        """

        :return: list of < video object > of trending music videos

        """

        url = f'https://www.youtube.com/feed/music'
        raw = urllib.request.urlopen(url).read().decode()
        data = re.findall(r"videoId\":\"(.*?)\"", raw)
        return _VideoBulk(_filter(data)) if len(data) > 0 else None


    @property
    def Gaming(self):

        """

        :return: list of < video object > of trending gaming videos

        """

        url = f'https://www.youtube.com/gaming'
        raw = urllib.request.urlopen(url).read().decode()
        data = re.findall(r"videoId\":\"(.*?)\"", raw)
        return _VideoBulk(_filter(data)) if len(data) > 0 else None


    @property
    def News(self):

        """

        :return: list of < video object > of trending news videos

        """

        url = f'https://www.youtube.com/news'
        raw = urllib.request.urlopen(url).read().decode()
        data = re.findall(r"videoId\":\"(.*?)\"", raw)
        return _VideoBulk(_filter(data)) if len(data) > 0 else None

    @property
    def Live(self):

        """

        :return: list of < video object > of trending livestreams

        """

        url = f'https://www.youtube.com/live'
        raw = urllib.request.urlopen(url).read().decode()
        data = re.findall(r"videoId\":\"(.*?)\"", raw)
        return _VideoBulk(_filter(data)) if len(data) > 0 else None

    @property
    def Learning(self):

        """

        :return: list of < video object > of trending educational videos

        """

        url = f'https://www.youtube.com/learning'
        raw = urllib.request.urlopen(url).read().decode()
        data = re.findall(r"videoId\":\"(.*?)\"", raw)
        return _VideoBulk(_filter(data)) if len(data) > 0 else None


    @property
    def Sports(self):

        """

        :return: list of < video object > of trending sports videos

        """
        url = f'https://www.youtube.com/sports'
        raw = urllib.request.urlopen(url).read().decode()
        data = re.findall(r"videoId\":\"(.*?)\"", raw)
        return _VideoBulk(_filter(data)) if len(data) > 0 else None


import re
import json
import urllib.request
from .__vid__ import Video
from .__proc__ import _filter
from .__hyp__ import _HyperThread
from urllib.error import HTTPError
from .__vidbulk__ import _VideoBulk
from .__pllsbulk__ import _PlaylistBulk



class Channel:

    def __init__(self, channelId:str):

        """
        :param str channelId: any of channel _id, _url , custom _url
        """

        if '/channel/' in channelId:
            self._url = channelId

        elif '/c/' in channelId:
            self._url = channelId

        elif '/user/' in channelId:
            self._url = channelId
        else:
            self._url = f'https://www.youtube.com/channel/{channelId}'



    def __repr__(self):
        _name = self.name
        return _name if _name is not None else 'invalid'


    @property
    def valid(self):
        _url = self._url
        try:
            urllib.request.urlopen(_url)
            return True
        except HTTPError as e:
            if e.code == 404:
                return False
            elif e.code == 429:
                return None


    @property
    def url(self):
        return self._url


    @property
    def id(self):

        """
        :return: the ID of the channel
        """

        raw = urllib.request.urlopen(f'{self._url}/about').read().decode()
        Id = re.search(r"\"channelId\":\"(.*?)\"", raw)
        return Id.group().replace('"','').replace('channelId:','') if Id else None



    @property
    def verified(self):

        """
        :return: bool i.e. True if channel is verified else False
        """

        raw = urllib.request.urlopen(f'{self._url}/about').read().decode()
        isVerified = re.search(r'label":"Verified', raw)
        return True if isVerified else False


    @property
    def live(self):

        """
        :return: Bool of channel's Live Status
        """

        raw = urllib.request.urlopen(f'{self._url}/videos?view=2&live_view=501').read().decode()
        return '{"text":" watching"}' in raw



    @property
    def streaming_now(self):

        """
        :return: channel's ongoing  livestream url
        """

        raw = urllib.request.urlopen(f'{self._url}/videos?view=2&live_view=501').read().decode()

        if '{"text":" watching"}' in raw:
            Id = _filter(re.findall(r"\"videoId\":\"(.*?)\"", raw))[0]
            return f'https://www.youtube.com/watch?v={Id}'



    @property
    def streaming_all_now(self):

        """
        :return: channel's ongoing  livestream urls
        """

        raw = urllib.request.urlopen(
            f'{self._url}/videos?view=2&live_view=501'
        ).read().decode()

        if '{"text":" watching"}' in raw:
            Ids = _filter(re.findall(r"\"videoId\":\"(.*?)\"", raw))
            return [f'https://www.youtube.com/watch?v={Id}' for Id in Ids]



    @property
    def past_streams(self):

        """
        :return: channel's old livestream urls
        """

        raw = urllib.request.urlopen(
            f'{self._url}/videos?view=2&live_view=503'
        ).read().decode()

        Ids = _filter(re.findall(r"\"videoId\":\"(.*?)\"", raw))
        urls = [f'https://www.youtube.com/watch?v={Id}' for Id in Ids]
        return urls if len(urls) > 0 else None



    def uploads(self, limit:int = None):

        """
        :param int limit: number of videos user wants from channel's latest upload
        :return: a < bulk video obj > of latest uploaded videos (consider limit)
        """

        query = f'{self._url}/videos?view=0&sort=dd&flow=grid'
        raw = urllib.request.urlopen(query).read().decode()
        videos = re.findall(r"\"gridVideoRenderer\":{\"videoId\":\"(.*?)\"", raw)
        limited = videos[:(limit - len(videos))] if limit is not None and limit < 30 else videos
        return _VideoBulk(limited) if len(limited) > 0 else None



    @property
    def latest(self):

        """
        :return: Channel's latest uploaded video in Video Object form
        """
        def check(Id):
            if 'live' not in Video(Id).thumbnail:
                return Id
            else:
                return None

        query = f'{self._url}/videos?view=0&sort=dd&flow=grid'
        raw = urllib.request.urlopen(query).read().decode()
        videos = re.findall(r"gridVideoRenderer\":{\"videoId\":\"(.*?)\"", raw)
        if len(videos) > 0:
            if '_live' not in raw:
                return Video(videos[0])
            else:
                ls = _HyperThread.run(check,videos)
                true_type = [Id for Id in ls if Id is not None]
                return Video(true_type[0]) if len(true_type) > 0 else None



    @property
    def info(self):

        """
        :return: a dict containing channel info

        dict = {
            'name': -> str,
            'id': -> str,
            'subscribers': -> str,
            'verified': -> bool,
            'total_views': -> str,
            'joined_at': -> str,
            'country': -> str,
            'url': ->str,
            'custom_url': -> str,
            'avatar_url': -> str,
            'banner_url': -> str
            }

        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()

        def get_data(pattern):
            data = re.findall(pattern, raw)
            return data[0] if len(data) > 0 else None

        patterns = [

            "channelMetadataRenderer\":{\"title\":\"(.*?)\"",
            "subscriberCountText\":(.*?)B",
            "\"viewCountText\":{\"simpleText\":\"(.*?)\"}",
            "{\"text\":\"Joined \"},{\"text\":\"(.*?)\"}",
            "\"country\":{\"simpleText\":\"(.*?)\"}",
            "\"canonicalChannelUrl\":\"(.*?)\"",
            "height\":88},{\"url\":\"(.*?)\"",
            "width\":1280,\"height\":351},{\"url\":\"(.*?)\"",
            "channelId\":\"(.*?)\""

        ]

        ls = _HyperThread.run(get_data, patterns)
        
        if len(ls[1]) > 0:
            strJSON = ls[1][0].replace(',"tv', '')
            dct = json.loads(strJSON)
            sub = dct['simpleText'].split(' ')[0]
        else:
            sub = None
            
        if len(ls[2]) > 0:
            views = ls[2].split(' ')[0]
        else:
            views = None

        infoDict = {
            'name': ls[0],
            'id': ls[8],
            'subscribers': sub,
            'verified':self.verified,
            'total_views': views,
            'joined_at': ls[3],
            'country': ls[4],
            'url': self._url,
            'custom_url': ls[5],
            'avatar_url': ls[6],
            'banner_url': ls[7]
        }

        return infoDict


    @property
    def name(self):

        """
        :return: name of the channel or None
        """

        name_raw = urllib.request.urlopen(f'{self._url}/about').read().decode()
        titleList = re.findall(r"channelMetadataRenderer\":{\"title\":\"(.*?)\"", name_raw)
        return titleList[0] if len(titleList) > 0 else None


    @property
    def subscribers(self):

        """
        :return: total number of subscribers the channel has or None
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        dict_raw = re.findall("\"subscriberCountText\":(.*?)B", raw)
        if len(dict_raw) > 0:
            strJSON = dict_raw[0].replace(',"tv', '')
            final_dict = json.loads(strJSON)
            return final_dict['simpleText'].split(' ')[0]


    @property
    def total_views(self):

        """
        :return: total number of views the channel got or None
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        viewList = re.findall(r"\"viewCountText\":{\"simpleText\":\"(.*?)\"}", raw)
        if len(viewList) > 0:
            final = viewList[0].split(' ')
            return final[0]


    @property
    def joined(self):

        """
        :return: the channel creation date or None
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        joinList = re.findall(r"{\"text\":\"Joined \"},{\"text\":\"(.*?)\"}", raw)
        return joinList[0] if len(joinList) != 0 else None


    @property
    def country(self):

        """
        :return: the name of the country from where the channel is or None
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        countryList = re.findall(r"\"country\":{\"simpleText\":\"(.*?)\"}", raw)
        return countryList[0] if len(countryList) > 0 else None


    @property
    def custom_url(self):

        """
        :return: the custom _url of the channel or None
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        customList = re.findall(r"\"canonicalChannelUrl\":\"(.*?)\"", raw)
        customURL = customList[0] if len(customList) != 0 else None
        return customURL if '/channel/' not in customURL and customURL is not None else None


    @property
    def description(self):

        """
        :return: the existing description of the channel
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        descList = re.findall(r"{\"description\":{\"simpleText\":\"(.*?)\"}", raw)
        return descList[0].replace('\\n', '  ') if len(descList) > 0 else None


    @property
    def avatar_url(self):

        """
        :return: logo / avatar url of the channel
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        data = re.findall("height\":88},{\"url\":\"(.*?)\"",raw)
        return data[0] if len(data) != 0 else None


    @property
    def banner_url(self):

        """
        :return: banner url of the channel
        """

        query = f'{self._url}/about'
        raw = urllib.request.urlopen(query).read().decode()
        data = re.findall(r"width\":1280,\"height\":351},{\"url\":\"(.*?)\"",raw)
        return data[0] if len(data) != 0 else None


    @property
    def playlists(self):

        """
        :return: a list of < playlist object > for each public playlist the channel has
        """

        url = f'{self._url}/playlists'
        raw = urllib.request.urlopen(url).read().decode()
        idList = re.findall(r"{\"url\":\"/playlist\?list=(.*?)\"", raw)
        return _PlaylistBulk(_filter(idList)) if len(idList) > 0 else None



from pytube import YouTube
from collections import OrderedDict


def _filter(iterable:list, limit:int = None):
    """
    Restricts element repetition in iterable

    :param int limit: number of desired elements
    :param iterable: list or tuple of elements
    :return: modified list (consider limit)

    """
    lim = limit if limit is not None else 0
    converted = list(OrderedDict.fromkeys(iterable))
    return converted[:-len(converted) + lim] if len(converted) > len(converted) - lim > 0 else converted


def _duration(seconds: int):

    """
    :param seconds: duration to be converted
    :return: a duration string with 00h 00m 00s format

    """

    dur_hour = int(seconds // 3600)
    dur_min = int((seconds % 3600) // 60)
    dur_sec = int(seconds - (3600 * dur_hour) - (60 * dur_min))
    return f'{dur_hour}h {dur_min}m {dur_sec}s'


def _parser(kw:str):
    query = kw.replace(" ", '+')
    return query


def _astream(Id: str):
    url = f'https://www.youtube.com/watch?v={Id}'
    temp = YouTube(url)
    stream = temp.streaming_data['formats']
    for item in stream:
        if item['itag'] == 17:
            return item['url']


def _vstream(Id: str):
    url = f'https://www.youtube.com/watch?v={Id}'
    temp = YouTube(url)
    stream = temp.streaming_data['formats']
    for item in stream:
        if item['itag'] == 22:
            return item['url']



import concurrent.futures


class _HyperThread:

    def __init__(self):
        pass


    @classmethod
    def run(cls, function, args: list) -> list:
        """
        Applicable for single-func 1-arg each thread
        :param function: the single-agred parent func to be executed
        :param args: list of arg
        :return: whole list of the values returned by the parent func in each thread
        """

        num = len(args) if len(args) > 0 else 1
        if len(args) > 0:
            with concurrent.futures.ThreadPoolExecutor(max_workers=num) as exec:
                res = list(exec.map(function, args))
            return res

        



from setuptools import setup

def readme():
    with open('README.md') as file:
        README = file.read()
    return README


setup(
    name = 'aiotube',
    version = '1.2.4',
    description = 'Get YouTube Public Data without YouTubeAPI',
    long_description = readme(),
    long_description_content_type="text/markdown",
    package_dir={'aiotube': 'src'},
    packages=['aiotube'],
    install_requires = [
        'urllib3','pytube'
    ],
    url = 'https://github.com/jnsougata/AioTube/blob/main/README.md',
    project_urls={
        "Bug Tracker": "https://github.com/jnsougata/AioTube/issues"
    },
    author = 'Sougata Jana',
    author_email = 'jnsougata@gmail.com',
    license = 'MIT',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    python_requires=">=3.6"
)



import re
import urllib.request as req
from .__hyp__ import _HyperThread



class _PlaylistBulk:

    def __init__(self, iterable: list):
        self._ls = iterable

    @property
    def name(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"{\"title\":\"(.*?)\""
            return re.findall(pattern, raw)[0]

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/playlist?list={item}' for item in self._ls])


    @property
    def url(self):
        return [f'https://www.youtube.com/playlist?list={item}' for item in self._ls]

    @property
    def video_count(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"stats\":\[{\"runs\":\[{\"text\":\"(.*?)\""
            temp = re.findall(pattern, raw)
            return temp[0] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/playlist?list={item}' for item in self._ls])

    @property
    def thumbnail(self):
        def _get_data(url: str):
            raw = req.urlopen(url).read().decode()
            pattern = r"og:image\" content=\"(.*?)\?"
            temp = re.findall(pattern, raw)
            return temp[0] if len(temp) > 0 else None

        return _HyperThread.run(_get_data, [f'https://www.youtube.com/playlist?list={item}' for item in self._ls])


import re
import urllib.request
from .__vid__ import Video
from .__ch__ import Channel
from .__plls__ import Playlist
from .__vidbulk__ import _VideoBulk
from .__chbulk__ import _ChannelBulk
from .__proc__ import _parser, _filter
from .__pllsbulk__ import _PlaylistBulk


class Search:

    def __init__(self):
        pass



    @classmethod
    def video(cls, keywords:str):

        """

        :return: < video object > regarding the query

        """

        url = f'https://www.youtube.com/results?search_query={_parser(keywords)}&sp=EgIQAQ%253D%253D'
        raw = urllib.request.urlopen(url).read().decode()
        video_ids = re.findall(r"\"videoId\":\"(.*?)\"", raw)
        return Video(video_ids[0]) if len(video_ids) != 0 else None


    @classmethod
    def channel(cls, keywords:str):

        """

        :return: < channel object > regarding the query

        """

        url = f'https://www.youtube.com/results?search_query={_parser(keywords)}&sp=EgIQAg%253D%253D'
        raw = urllib.request.urlopen(url).read().decode()
        channel_ids = re.findall(r"{\"channelId\":\"(.*?)\"", raw)
        return Channel(channel_ids[0]) if len(channel_ids) != 0 else None


    @classmethod
    def videos(cls, keywords:str, limit: int = None):

        """
        :param str keywords: query to be searched on YouTube
        :param int limit: total number of videos to be searched
        :return: list of < video object > of each video regarding the query (consider limit)

        """

        url = f'https://www.youtube.com/results?search_query={_parser(keywords)}&sp=EgIQAQ%253D%253D'
        raw = urllib.request.urlopen(url).read().decode()
        raw_ids = re.findall(r"\"videoId\":\"(.*?)\"", raw)
        pureList = _filter(limit=limit, iterable=raw_ids)
        return _VideoBulk(pureList) if len(pureList) != 0 else None


    @classmethod
    def channels(cls, keywords:str, limit: int = None):

        """
        :param str keywords: query to be searched on YouTube
        :param int limit: total number of channels to be searched
        :return: list of < channel object > of each video regarding the query (consider limit)

        """

        url = f'https://www.youtube.com/results?search_query={_parser(keywords)}&sp=EgIQAg%253D%253D'
        raw = urllib.request.urlopen(url).read().decode()
        raw_ids = re.findall(r"{\"channelId\":\"(.*?)\"", raw)
        pureList = _filter(limit=limit, iterable=raw_ids)
        return _ChannelBulk(pureList) if len(pureList) != 0 else None


    @classmethod
    def playlist(cls, keywords:str):

        """

        :return: < playlist object > regarding the query

        """

        url = f'https://www.youtube.com/results?search_query={_parser(keywords)}&sp=EgIQAw%253D%253D'
        raw = urllib.request.urlopen(url=url).read().decode()
        found = re.findall(r"playlistId\":\"(.*?)\"", raw)
        return Playlist(found[0]) if len(found) != 0 else None


    @classmethod
    def playlists(cls, keywords:str, limit: int = None):

        """
        :param str keywords: query to be searched on YouTube
        :param int limit: total playlists be searched
        :return: list of < playlist object > of each playlist regarding the query (consider limit)

        """

        url = f'https://www.youtube.com/results?search_query={_parser(keywords)}&sp=EgIQAw%253D%253D'
        raw = urllib.request.urlopen(url=url).read().decode()
        found = re.findall(r"playlistId\":\"(.*?)\"", raw)
        pure = _filter(limit = limit, iterable = found)
        return _PlaylistBulk(pure) if len(pure) != 0 else None


import re
import requests
import urllib.request
from .__hyp__ import _HyperThread
from .__proc__ import _duration, _astream, _vstream

class Video:

    def __init__(self, videoId: str):

        """
        :param videoId: video id or the url of the video
        """

        if 'watch?v=' in videoId:
            self._url = videoId
            self._id = re.findall(r"v=(.*)", videoId)[0]

        elif 'youtu.be/' in videoId:
            idL = re.findall(r"youtu\.be/(.*)", videoId)
            self._url = f'https://www.youtube.com/watch?v={idL[0]}'
            self._id = re.findall(r"/(.*)", videoId)[0]
        else:
            self._url = f'https://www.youtube.com/watch?v={videoId}'
            self._id = videoId


    @property
    def url(self):
        return self._url


    @property
    def id(self):
        return self._id


    @property
    def title(self):

        """
        :return: the title of the video
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(r"\"title\":\"(.*?)\"", raw)
        return data[0] if len(data) > 0 else None


    @property
    def views(self):

        """
        :return: total views the video got so far
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(
            r"\"videoViewCountRenderer\":{\"viewCount\":{\"simpleText\":\"(.*?)\"",
            raw
        )
        return data[0][:-6] if len(data) > 0 else None


    @property
    def likes(self):

        """
        :return: total likes the video got so far
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(
            r"toggledText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?) ",
            raw
        )
        return data[0] if len(data) > 0 else None


    @property
    def dislikes(self):

        """
        :return: total dislikes the video got so far
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(
            r"DISLIKE\"},\"defaultText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?) dislikes\"",
            raw
        )
        return data[0] if len(data) > 0 else None


    @property
    def duration(self):

        """
        :return: total duration of  the video in 00h 00m 00s
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(r"approxDurationMs\":\"(.*?)\"", raw)

        return _duration(int(int(data[0]) / 1000)) if len(data) > 0 else None


    @property
    def upload_date(self):

        """
        :return: the date on which the video has been uploaded
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(r"uploadDate\":\"(.*?)\"", raw)
        return data[0] if len(data) > 0 else None


    @property
    def parent(self):

        """
        :return: the id of the channel from which the video belongs
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(r"channelIds\":\[\"(.*?)\"", raw)
        return data[0] if len(data) > 0 else None


    @property
    def description(self):

        """
        :return: description provided with the video
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(r"shortDescription\":\"(.*)\",\"isCrawlable", raw)
        return data[0].replace('\\n', '') if len(data) > 0 else None


    @property
    def thumbnail(self):

        """
        :return: _url of the thumbnail of the video
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(
            r"playerMicroformatRenderer\":{\"thumbnail\":{\"thumbnails\":\[{\"url\":\"(.*?)\"",
            raw
        )
        return data[0] if len(data) > 0 else None


    @property
    def tags(self):

        """
        :return: list of tags used in video meta-data
        """

        raw = urllib.request.urlopen(self._url).read().decode()
        data = re.findall(r"<meta name=\"keywords\" content=\"(.*?)\">", raw)
        return data[0].split(',') if len(data) > 0 else None


    @property
    def info(self):

        """
        :return: dict containing the the whole info of the video

        dict = {

            'title': -> str,
            'id': -> str,
            'views': -> str,
            'likes': -> str,
            'dislikes': -> str,
            'duration': -> str,
            'parent': -> str,
            'uploaded': -> str,
            'url': -> str,
            'thumbnail': -> str,
            'tags': -> list,
        }
        """

        raw = urllib.request.urlopen(self._url).read().decode()

        def _get_data(pattern):
            data = re.findall(pattern, raw)
            return data[0] if len(data) > 0 else None

        patterns = [

            r"\"title\":\"(.*?)\"",
            r"\"videoViewCountRenderer\":{\"viewCount\":{\"simpleText\":\"(.*?)\"",
            r"toggledText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?) ",
            r"DISLIKE\"},\"defaultText\":{\"accessibility\":{\"accessibilityData\":{\"label\":\"(.*?) dislikes\"",
            r"approxDurationMs\":\"(.*?)\"",
            r"channelIds\":\[\"(.*?)\"",
            r"uploadDate\":\"(.*?)\"",
            r"playerMicroformatRenderer\":{\"thumbnail\":{\"thumbnails\":\[{\"url\":\"(.*?)\"",
            r"<meta name=\"keywords\" content=\"(.*?)\">"

        ]

        ls = _HyperThread.run(_get_data, patterns)


        infoDict = {

            'title': ls[0],
            'id': self._id,
            'views': ls[1][:-6],
            'likes': ls[2],
            'dislikes': ls[3],
            'duration': _duration(int(int(ls[4]) / 1000)),
            'parent': ls[5],
            'uploaded': ls[6],
            'url': self._url,
            'thumbnail': ls[7],
            'tags': ls[8].split(','),

        }

        return infoDict


    def download(
            self,
            format:str,
            filename:str = None,
    ):

        if filename:
            prefix = filename.replace(' ','')
        else:
            prefix = 'aiofile'

        def check(ext):

            if ext == 'mp3':
                return _astream(self._id)

            elif ext == 'mp4':
                return _vstream(self._id)

            else:
                raise ValueError("invalid format. use mp3 or mp4")

        try:
            stream = check(ext=format)
            if stream:
                r = requests.get(stream, stream=True)

                with open(f"./download/{prefix}_{self._id}.mp3", "wb") as file:
                    for chunk in r.iter_content(chunk_size=512):
                        if chunk:
                            file.write(chunk)

                        print('Download completed!')
            else:
                raise RuntimeError('unable to retrieve file')
        except:
            raise RuntimeError('unable to download file')

